"""Resolwe annotations serializer."""
from django.db import models

from rest_framework import serializers

from resolwe.flow.models.annotations import (
    AnnotationField,
    AnnotationGroup,
    AnnotationPreset,
    AnnotationValue,
)

from .base import ResolweBaseSerializer
from .fields import PrimaryKeyDictRelatedField


class AnnotationFieldListSerializer(serializers.ListSerializer):
    """Override list serializer to allow filtering nested relation."""

    def to_representation(self, data: "models.QuerySet[AnnotationGroup]"):
        """Filter the annotation fields within the group by the collection."""
        if (
            "request" in self.context
            and "collection_id" in self.context["request"].query_params
        ):
            collection_id = self.context["request"].query_params["collection_id"]
            data = data.filter(collections__id=collection_id)
        return super().to_representation(data)


class AnnotationGroupSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationGroup objects."""

    class Meta:
        """AnnotationGroupSerializer Meta options."""

        model = AnnotationGroup
        read_only_fields = ("id", "label", "name", "sort_order")
        fields = read_only_fields


class AnnotationFieldSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationField objects."""

    group = AnnotationGroupSerializer()

    class Meta:
        """AnnotationFieldSerializer Meta options."""

        model = AnnotationField
        read_only_fields = ("id",)
        fields = read_only_fields + (
            "collection",
            "description",
            "group",
            "label",
            "name",
            "sort_order",
            "type",
            "validator_regex",
            "vocabulary",
        )
        list_serializer_class = AnnotationFieldListSerializer


class AnnotationFieldDictSerializer(serializers.Serializer):
    """Serializer for actions that deal with annotation fields."""

    annotation_fields = PrimaryKeyDictRelatedField(
        queryset=AnnotationField.objects.all(), required=True, many=True
    )
    confirm_action = serializers.BooleanField(default=False)


class AnnotationsSerializer(serializers.Serializer):
    """Serializer that reads annotation field and its value."""

    field = PrimaryKeyDictRelatedField(
        queryset=AnnotationField.objects.all(), required=True
    )
    value = serializers.JSONField()


class AnnotationPresetSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationPreset objects."""

    class Meta:
        """AnnotationPresetSerializer Meta options."""

        model = AnnotationPreset
        read_only_fields = ("id",)
        fields = read_only_fields + ("name", "fields", "contributor")


class AnnotationValueSerializer(ResolweBaseSerializer):
    """Serializer for AnnotationValue objects."""

    class Meta:
        """AnnotationValueSerializer Meta options."""

        model = AnnotationValue
        read_only_fields = ("id", "field")
        update_protected_fields = ("entity", "field")
        fields = read_only_fields + update_protected_fields + ("value",)



""".. Ignore pydocstyle D400.

==========================
Command: runstoragecleanup
==========================

"""
import logging

from asgiref.sync import async_to_sync
from channels.layers import ChannelFull, get_channel_layer

from django.core.management.base import BaseCommand

from resolwe.storage.protocol import (
    CHANNEL_STORAGE_CLEANUP_WORKER,
    TYPE_STORAGE_CLEANUP_RUN,
)

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    """Start storage cleanup run via signal by django channels."""

    help = "Start cleanup manager run via signal by django channels."

    def handle(self, *args, **options):
        """Command handle."""
        channel_layer = get_channel_layer()
        try:
            async_to_sync(channel_layer.send)(
                CHANNEL_STORAGE_CLEANUP_WORKER,
                {
                    "type": TYPE_STORAGE_CLEANUP_RUN,
                },
            )
        except ChannelFull:
            logger.warning(
                "Cannot trigger storage manager run because channel is full.",
            )



"""Implementation of field projection."""
from collections.abc import Mapping, Sequence

FIELD_SEPARATOR = ","
FIELD_DEREFERENCE = "__"


def apply_subfield_projection(field, value, deep=False):
    """Apply projection from request context.

    The passed dictionary may be mutated.

    :param field: An instance of `Field` or `Serializer`
    :type field: `Field` or `Serializer`
    :param value: Dictionary to apply the projection to
    :type value: dict
    :param deep: Also process all deep projections
    :type deep: bool
    """
    # Discover the root manually. We cannot use either `self.root` or `self.context`
    # due to a bug with incorrect caching (see DRF issue #5087).
    prefix = []
    root = field
    while root.parent is not None:
        # Skip anonymous serializers (e.g., intermediate ListSerializers).
        if root.field_name:
            prefix.append(root.field_name)
        root = root.parent
    prefix = prefix[::-1]

    context = getattr(root, "_context", {})

    # If there is no request, we cannot perform filtering.
    request = context.get("request")
    if request is None:
        return value

    filtered = set(request.query_params.get("fields", "").split(FIELD_SEPARATOR))
    filtered.discard("")
    if not filtered:
        # If there are no fields specified in the filter, return all fields.
        return value

    # Extract projection for current and deeper levels.
    current_level = len(prefix)
    current_projection = []
    for item in filtered:
        item = item.split(FIELD_DEREFERENCE)
        if len(item) <= current_level:
            continue

        if item[:current_level] == prefix:
            if deep:
                current_projection.append(item[current_level:])
            else:
                current_projection.append([item[current_level]])

    if deep and not current_projection:
        # For deep projections, an empty projection means that all fields should
        # be returned without any projection.
        return value

    # Apply projection.
    return apply_projection(current_projection, value)


def apply_projection(projection, value):
    """Apply projection."""
    if isinstance(value, Sequence):
        # Apply projection to each item in the list.
        return [apply_projection(projection, item) for item in value]
    elif not isinstance(value, Mapping):
        # Non-dictionary values are simply ignored.
        return value

    # Extract projection for current level.
    try:
        current_projection = [p[0] for p in projection]
    except IndexError:
        return value

    # Apply projection.
    for name in list(value.keys()):
        if name not in current_projection:
            value.pop(name)
        elif isinstance(value[name], dict):
            # Apply projection recursively.
            value[name] = apply_projection(
                [p[1:] for p in projection if p[0] == name], value[name]
            )

    return value



"""Decorators for observers."""


from contextlib import suppress

from resolwe.observers.models import Observable
from resolwe.observers.protocol import (
    post_container_changed,
    pre_container_changed,
    suppress_notifications_attribute,
)


def move_to_container(function):
    """Indicate object is moving between containers.

    The notifications are suppressed for the first argument (usually self) of the given
    function which must be the observable object.
    """

    def wrapper(self, *args, **kwargs):
        pre_container_changed.send(sender=type(self), instance=self)
        suppress_observer_notifications(function)(self, *args, **kwargs)
        post_container_changed.send(sender=type(self), instance=self)

    return wrapper


def suppress_observer_notifications(function):
    """Suppress the observer notifications.

    The notifications are suppressed for the first argument (usually self) of the given
    function which must be the observable object.

    :raises AssertionError: when first argument of the given function is not observable
        object.
    """

    def wrapper(self, *args, **kwargs):
        assert isinstance(
            self, Observable
        ), "The first argument of the method must be observable object."
        setattr(self, suppress_notifications_attribute, True)
        try:
            function(self, *args, **kwargs)
        finally:
            with suppress(AttributeError):
                delattr(self, suppress_notifications_attribute)

    return wrapper



"""An execution engine that supports workflow specifications."""
import collections

import yaml

from django.db import transaction

from resolwe.flow.execution_engines.base import BaseExecutionEngine
from resolwe.flow.execution_engines.exceptions import ExecutionError
from resolwe.flow.expression_engines import EvaluationError
from resolwe.flow.models import Data, Process


class ExecutionEngine(BaseExecutionEngine):
    """An execution engine that supports workflow specifications."""

    name = "workflow"

    def discover_process(self, path):
        """Perform process discovery in given path.

        This method will be called during process registration and
        should return a list of dictionaries with discovered process
        schemas.
        """
        if not path.lower().endswith((".yml", ".yaml")):
            return []

        with open(path) as fn:
            schemas = yaml.load(fn, Loader=yaml.FullLoader)
        if not schemas:
            # TODO: Logger.
            # self.stderr.write("Could not read YAML file {}".format(schema_file))
            return []

        process_schemas = []
        for schema in schemas:
            if "run" not in schema:
                continue

            # NOTE: This currently assumes that 'bash' is the default.
            if schema["run"].get("language", "bash") != "workflow":
                continue

            process_schemas.append(schema)

        return process_schemas

    def get_output_schema(self, process):
        """Return any additional output schema for the process."""
        return [
            {"name": "steps", "label": "Steps", "type": "list:data:"},
        ]

    def _evaluate_expressions(self, expression_engine, step_id, values, context):
        """Recursively evaluate expressions in a dictionary of values."""
        if expression_engine is None:
            return values

        processed = {}
        for name, value in values.items():
            if isinstance(value, str):
                value = value.strip()
                try:
                    expression = expression_engine.get_inline_expression(value)
                    if expression is not None:
                        # Inline expression.
                        value = expression_engine.evaluate_inline(expression, context)
                    else:
                        # Block expression.
                        value = expression_engine.evaluate_block(value, context)
                except EvaluationError as error:
                    raise ExecutionError(
                        'Error while evaluating expression for step "{}":\n{}'.format(
                            step_id, error
                        )
                    )
            elif isinstance(value, dict):
                value = self._evaluate_expressions(
                    expression_engine, step_id, value, context
                )

            if value is not None:
                processed[name] = value

        return processed

    @transaction.atomic
    def evaluate(self, data):
        """Evaluate the code needed to compute a given Data object."""
        expression_engine = data.process.requirements.get("expression-engine", None)
        if expression_engine is not None:
            expression_engine = self.get_expression_engine(expression_engine)

        # Parse steps.
        steps = data.process.run.get("program", None)
        if steps is None:
            return

        if not isinstance(steps, list):
            raise ExecutionError("Workflow program must be a list of steps.")

        # Expression engine evaluation context.
        context = {
            "input": data.input,
            "steps": collections.OrderedDict(),
        }

        for index, step in enumerate(steps):
            try:
                step_id = step["id"]
                step_slug = step["run"]
            except KeyError as error:
                raise ExecutionError(
                    'Incorrect definition of step "{}", missing property "{}".'.format(
                        step.get("id", index), error
                    )
                )

            # Fetch target process.
            process = (
                Process.objects.filter(slug=step_slug).order_by("-version").first()
            )
            if not process:
                raise ExecutionError(
                    'Incorrect definition of step "{}", invalid process "{}".'.format(
                        step_id, step_slug
                    )
                )

            # Process all input variables.
            step_input = step.get("input", {})
            if not isinstance(step_input, dict):
                raise ExecutionError(
                    'Incorrect definition of step "{}", input must be a dictionary.'.format(
                        step_id
                    )
                )

            data_input = self._evaluate_expressions(
                expression_engine, step_id, step_input, context
            )

            # Create the data object.
            data_object = Data.objects.create(
                process=process,
                contributor=data.contributor,
                tags=data.tags,
                input=data_input,
                collection=data.collection,
                subprocess_parent=data,
            )

            context["steps"][step_id] = data_object.pk

        # Immediately set our status to done and output all data object identifiers.
        data.output = {
            "steps": list(context["steps"].values()),
        }
        data.status = Data.STATUS_DONE



"""Local Storage connector."""
import os
import shutil
from pathlib import Path

from .baseconnector import (
    BaseStorageConnector,
    ConnectorType,
    validate_url,
    validate_urls,
)
from .hasher import StreamHasher


class LocalFilesystemConnector(BaseStorageConnector):
    """Local filesystem connector."""

    #: Read files by chunks of the given size
    REQUIRED_SETTINGS = ["path"]
    CONNECTOR_TYPE = ConnectorType.LOCAL

    def __init__(self, config: dict, name: str):
        """Connector initialization."""
        super().__init__(config, name)
        self.path = config["path"]
        self.supported_hash = ["crc32c", "md5", "awss3etag"]
        self.multipart_chunksize = self.CHUNK_SIZE
        self.get_ensures_data_integrity = True
        self.put_ensures_data_integrity = True

    @property
    def mountable(self):
        """Can a connector be mounted inside container."""
        return True

    def prepare_url(self, url, **kwargs):
        """Prepare all the necessary for a new location."""
        (self.base_path / url).mkdir(mode=kwargs.get("dir_mode", 0o755), parents=True)

    @validate_urls
    @validate_url
    def delete(self, url, urls):
        """Remove objects."""
        for delete_url in urls:
            path = self.base_path / url / delete_url
            if path.exists():
                if path.is_dir():
                    shutil.rmtree(os.fspath(path))
                else:
                    path.unlink()
        # Remove url base directory if empty.
        if not self.get_object_list(url):
            if (self.base_path / url).is_dir():
                shutil.rmtree(os.fspath(self.base_path / url))

    @validate_url
    def push(self, stream, url, chunk_size=BaseStorageConnector.CHUNK_SIZE, hashes={}):
        """Push data from the stream to the given URL.

        The chunk_size and hashes arguments are ignored.
        """
        data_remaining = True
        path = self.base_path / url
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("wb", chunk_size) as f:
            while data_remaining:
                data = stream.read(chunk_size)
                f.write(data)
                data_remaining = len(data) == chunk_size

    def multipart_push_start(self, url, size=None):
        """Start a multipart upload.

        :returns: the upload id.
        :raises AssertionError: when url exists or upload size is not given.
        """
        assert size is not None, f"{self}: size of multipart upload must be specified."
        path = self.base_path / url
        assert not path.exists(), f"Path {url} already exists, aborting upload."
        if not path.is_file():
            with path.open("wb") as stream:
                stream.truncate(size)
        return url

    def multipart_push(self, upload_id, url, part_number, chunk_size, data, md5=None):
        """Upload a single part of multipart upload."""
        path = self.base_path / url
        assert path.is_file(), f"{self}: multipart upload file {path} does not exist."
        with path.open("r+b") as stream:
            stream.seek((part_number - 1) * chunk_size)
            shutil.copyfileobj(data, stream, 1024 * 1024)
        return dict()

    def multipart_push_complete(self, upload_id, url, completed_chunks):
        """Complete the multipart push."""
        path = self.base_path / url
        assert path.is_file(), f"{self}: multipart upload file {path} does not exist."
        return dict()

    def multipart_push_abort(self, upload_id, url):
        """Abort multiport upload."""
        path = self.base_path / url
        if path.is_file():
            path.unlink()

    @validate_url
    def get(self, url, stream, chunk_size=BaseStorageConnector.CHUNK_SIZE):
        """Get data from the given URL and write it into the given stream."""
        path = self.base_path / url
        with path.open("rb", chunk_size) as f:
            for chunk in iter(lambda: f.read(chunk_size), b""):
                stream.write(chunk)

    @property
    def can_open_stream(self):
        """Get True if connector can open object as stream."""
        return True

    @validate_url
    def open_stream(self, url, mode):
        """Get stream for data at the given URL."""
        path: Path = self.base_path / url
        path.parent.mkdir(parents=True, exist_ok=True)
        return path.open(mode, self.CHUNK_SIZE)

    @validate_url
    def get_object_list(self, url):
        """Get a list of objects stored bellow the given URL."""
        path = self.base_path / url
        return [
            os.fspath((Path(dirpath) / filename).relative_to(path))
            for dirpath, _, files in os.walk(path)
            for filename in files
            if filename != path
        ]

    @validate_url
    def exists(self, url):
        """Get if the object at the given URL exist."""
        return (self.base_path / url).exists()

    @validate_url
    def get_hashes(self, url, hash_types):
        """Get the hash of the given type for the given object."""
        hasher = StreamHasher(chunk_size=self.multipart_chunksize, hashes=hash_types)
        path = self.base_path / url
        if not path.exists():
            return None
        with path.open("rb", self.CHUNK_SIZE) as f:
            hasher.compute(f)
        return {hash_type: hasher.hexdigest(hash_type) for hash_type in hash_types}

    @validate_url
    def get_hash(self, url, hash_type):
        """Get the hash of the given type for the given object."""
        hasher = StreamHasher(chunk_size=self.multipart_chunksize, hashes=[hash_type])
        path = self.base_path / url
        if not path.exists():
            return None
        with path.open("rb", self.CHUNK_SIZE) as f:
            hasher.compute(f)
        return hasher.hexdigest(hash_type)

    @validate_url
    def set_hashes(self, url, hashes):
        """Set the  hashes for the given object.

        Local connector has currently no way to store metadata alongside files
        so this call is silently ignored. Hash is calculated always when
        get_hash(es) method is called.
        """

    @property
    def base_path(self):
        """Get a base path for this connector."""
        return Path(self.path)

    @validate_url
    def url(self, url):
        """Return the URL to the given file in connector native notation.

        For local connector this is the same as the presigned URL.
        """
        return self.presigned_url(url)

    @validate_url
    def presigned_url(self, url, expiration=3600, force_download=False):
        """Create a presigned URL.

        The URL is used to obtain temporary access to the object ar the
        given URL using only returned URL.

        :param expiration: expiration time of the link (in seconds), default
            is one minute.

        :param force_download: force download.

        :returns: URL that can be used to access object or None.
        """
        force_download = "?force_download=1" if force_download else ""
        public_url = Path(self.config.get("public_url", ""))
        resource_url = public_url / url
        return resource_url.as_posix() + force_download



""".. Ignore pydocstyle D400.

===============================
Command: compare_models_and_csv
===============================

Usage: ./manage.py compare_models_and_csv <path_to_csv>

Check if all file references in django models (ReferencedPaths) correspond 1:1
to existing files in the remote database (using a CSV inventory file).
"""
import csv
import os
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, TextIO, Tuple

from django.core.management.base import BaseCommand, CommandParser
from django.db.models import Q
from django.db.models.query import QuerySet

from resolwe.flow.models import Data
from resolwe.storage.models import ReferencedPath


class Command(BaseCommand):
    """Compare Django's Data records with a CSV file and log differences.

    Expected CSV format:
        bucket_id, file_key, file_size, file_hash
    """

    help = (
        "Log differences between Data records in models (Django database)"
        " and records from a CSV file (S3 inventory)."
        " Expected CSV format: bucket_id, file_key, file_size, file_hash"
    )

    def add_arguments(self, parser: CommandParser) -> None:
        """Add command arguments."""
        # Positional argument: path to CSV file
        parser.add_argument("csv_path", nargs=1, type=str)

    def handle(self, *args: None, **options: Dict[Any, Any]) -> None:
        """Command handle."""
        csv: FileIterator = FileIterator(options["csv_path"][0])
        datas = Data.objects.all().order_by("location__pk")
        counter: Dict[str, int] = {
            "match": 0,
            "csv_only": 0,
            "models_only": 0,
            "hash_mismatch": 0,
        }
        subpath_map: SubpathMap = map_subpath_locations(csv)

        for data in datas.iterator():
            subpath = data.location.subpath
            urls = data.location.files
            urls = urls.exclude(Q(path__endswith="/"))  # exclude directories
            urls = ModelIterator(urls.order_by("path"))

            if subpath not in subpath_map:
                filecount = urls.count
                files = f"({filecount} files)"
                self.stdout.write(f"MODEL-ONLY {subpath}/* {files}")
                counter["models_only"] += filecount
                continue
            subpath_map[subpath]["visited"] = True

            csv.restrict(
                start=subpath_map[subpath]["start"],
                end=subpath_map[subpath]["end"],
            )
            csv.seek_relative(0)

            next_in_models, model_hash = urls.next()
            next_in_csv, csv_hash = csv.next()
            while next_in_csv and next_in_models:
                if next_in_models == next_in_csv:
                    # entries match, verify checksum
                    if model_hash == csv_hash:
                        counter["match"] += 1
                    else:
                        fullpath = f"{subpath}/{next_in_models}"
                        hashes = f"{model_hash} != {csv_hash}"
                        self.stdout.write(f"HASH {fullpath} {hashes}")
                        counter["hash_mismatch"] += 1
                    # advance both
                    next_in_models, model_hash = urls.next()
                    next_in_csv, csv_hash = csv.next()
                elif next_in_models < next_in_csv or not csv.has_next():
                    # entries are missing in CSV
                    # (models are alphabetically *behind*)
                    fullpath = subpath + "/" + next_in_models
                    self.stdout.write(f"MODEL-ONLY {fullpath}")
                    counter["models_only"] += 1
                    next_in_models, model_hash = urls.next()  # advance models
                elif next_in_models > next_in_csv or not urls.has_next():
                    # entries are missing in models
                    # (models are alphabetically *ahead*)
                    fullpath = subpath + "/" + next_in_csv
                    self.stdout.write(f"CSV-ONLY {fullpath}")
                    counter["csv_only"] += 1
                    next_in_csv, csv_hash = csv.next()  # advance CSV

            # either (or both) of the iterators is finished,
            # now we need to exhaust the other
            while next_in_csv:
                self.stdout.write(f"CSV-ONLY {subpath}/{next_in_csv}")
                counter["csv_only"] += 1
                next_in_csv, csv_hash = csv.next()
            while next_in_models:
                self.stdout.write(f"MODEL-ONLY {subpath}/{next_in_models}")
                counter["models_only"] += 1
                next_in_models, model_hash = urls.next()

        # list all subpaths from CSV that we haven't visited
        # while traversing models' data
        for subpath in subpath_map:
            if "visited" not in subpath_map[subpath]:
                filecount = subpath_map[subpath]["linecount"]
                self.stdout.write(f"CSV-ONLY {subpath}/* ({filecount} files)")
                counter["csv_only"] += filecount

        # print an overview/summary
        out = ""
        out += f"{counter['match']} files OK"
        if counter["csv_only"] != 0:
            out += f", {counter['csv_only']} files in CSV only"
        if counter["models_only"] != 0:
            out += f", {counter['models_only']} files in models only"
        if counter["hash_mismatch"] != 0:
            out += f", {counter['hash_mismatch']} files do not match the hash"
        self.stdout.write(out)

        # double check the numbers just in case
        ReferencedPath_count = ReferencedPath.objects.exclude(
            Q(path__endswith="/")
        ).count()
        self.stdout.write(f"CSV length = {csv.length}")
        self.stdout.write(f"ReferencedPath count = {ReferencedPath_count}")

        matches = counter["hash_mismatch"] + counter["match"]
        csv_records = matches + counter["csv_only"]
        models_records = matches + counter["models_only"]
        # this should never happen, but it's better to check,
        # just because it's so easy to do
        if csv_records != csv.length:
            self.stdout.write(
                "Numbers don't add up."
                " OK + csv_only + hash_mismatch != CSV.line_count."
            )
        if models_records != ReferencedPath_count:
            self.stdout.write(
                "Numbers don't add up."
                " OK + models_only + hash_mismatch != ReferencedPath_count."
                " There might be orphaned ReferencedPaths."
            )


Line = NamedTuple("Line", [("subpath", str), ("filename", str), ("hash", str)])


def parse_line(line: str) -> Line:
    """Parse a line of CSV data into an array of values."""

    reader = csv.reader([line])
    # this loop will only run once (because we supplied one line)
    for i in reader:
        fields = i

    file_key = fields[1]
    awss3etag = fields[3]

    # append an empty string to avoid an IndexError when file_key is ""
    parts = [part for part in Path(file_key).parts] + [""]
    subpath = str(parts[0])
    filename = str(Path(*parts[1:]))

    # edge case: str(Path("")) == "."
    # WRONG:
    # readme.md => subpath="readme.md", filename="."
    # "" => subpath=".", filename="."
    # RIGHT:
    # readme.md => subpath="", filename="readme.md"
    # "" => subpath="", filename=""
    if subpath == ".":
        # empty file key (parts[0] == "")
        subpath = ""
    if filename == ".":
        # no subpath is present (ex. "README.md")
        # len(parts) == 1, so *parts[1:] == ""
        filename = subpath
        subpath = ""

    return Line(subpath=subpath, filename=filename, hash=awss3etag)


class FileIterator:
    """An iterator for reading a CSV file within a given range."""

    def __init__(self, csv_filename: str) -> None:
        """Initialize the FileIterator.

        :param csv_filename: The path to the CSV file to be read
        """
        self.name: str = csv_filename
        self.file: TextIO = open(self.name, "r")
        self.size: int = os.fstat(self.file.fileno()).st_size
        self.length: int = -1  # number of lines (must be set externally)
        self.last_position: int = 0
        self.restriction: Dict[str, int] = {
            "start": 0,
            "end": self.size,
        }

    def __del__(self) -> None:
        """Close the file."""
        self.file.close()

    def tell(self) -> int:
        """Return the current position of the reader."""
        return self.file.tell()

    def seek(self, position: int) -> None:
        """Seek to a given position within the file.

        This may position the reader outside the restriction interval,
        resulting in has_next() always returning False.
        """
        self.last_position = self.tell()
        self.file.seek(position)

    def seek_relative(self, position: int) -> None:
        """Seek to a given position relative to the restriction interval."""
        self.seek(self.restriction["start"] + position)

    def readline(self) -> str:
        """Return the next line of CSV data."""
        self.last_position = self.tell()
        return self.file.readline()

    def next(self) -> Tuple[str, str]:
        """Return the next line's path string and etag.

        If the file has been exhausted (or reached the end of its restriction
        interval), ("", "") is returned.
        """
        if not self.has_next():
            return "", ""

        line = parse_line(self.readline())
        return line.filename, line.hash

    def has_next(self) -> bool:
        """Check if the file reached the end of its restriction interval."""
        return self.tell() < self.restriction["end"]

    def restrict(self, start: int = 0, end: int = -1) -> None:
        """Restrict reading to an interval from start to end.

        Call with default parameters to unrestrict.
        """
        if end == -1:
            end = self.size
        self.restriction = {"start": start, "end": end}


class ModelIterator:
    """An iterator for traversing a QuerySet.

    This is a wrapper with similar methods as the FileIterator, to provide
    nice symmetric-looking code.
    """

    def __init__(self, urls: QuerySet) -> None:
        """Initialize the ModelIterator.

        :param urls: A QuerySet of ReferencedPaths, ordered by path.
        """
        self.count = urls.count()

        self.urls: List[ReferencedPath] = [u for u in urls]
        # these should already be *mostly* sorted
        # problem is, SQL sorts "_" before ".",
        # but python does it the other way around
        # so this has to be re-sorted to match the sorting used within the CSV
        self.urls.sort(key=lambda a: a.path)

        self.i = 0

    def next(self) -> Tuple[str, str]:
        """Return the next ReferencedPath's path string and the file's etag.

        If all ReferencedPaths have been exhausted, ("", "") is returned.
        """
        if not self.has_next():
            return "", ""

        next_url = self.urls[self.i]
        self.i += 1

        return next_url.path, next_url.awss3etag

    def has_next(self) -> bool:
        """Check if all ReferencedPaths have been exhausted."""
        return self.i < self.count


SubpathData = Dict[str, int]
SubpathMap = Dict[str, SubpathData]


def map_subpath_locations(file: FileIterator) -> SubpathMap:
    """Map subpaths to their locations in a CSV file.

    Also count the number of lines for each subpath and the file as a whole.

    Sample output:
    ```
    {
        '1': {'start': 0, 'end': 234, 'linecount': 2},
        '100': {'start': 234, 'end': 876516, 'linecount': 7202},
        '101': {'start': 876516, 'end': 877441, 'linecount': 6},
        '86': {'start': 6324115, 'end': 6326268, 'linecount': 17},
        ...
        '94': {'start': 6338834, 'end': 7177568, 'linecount': 6975},
        '98': {'start': 7183154, 'end': 7184069, 'linecount': 6},
        '': {'start': 7184944, 'end': 7185044, 'linecount': 1}
    }
    ```
    """
    mapping: SubpathMap = {}
    last_subpath: Optional[str] = None
    current_subpath_linecount = 0
    file.seek(0)
    total_linecount = 0
    while file.has_next():
        subpath = parse_line(file.readline()).subpath
        if subpath != last_subpath:
            mapping[subpath] = {}
            mapping[subpath]["start"] = file.last_position
            if last_subpath is not None:
                mapping[last_subpath]["end"] = file.last_position
                mapping[last_subpath]["linecount"] = current_subpath_linecount
            last_subpath = subpath
            current_subpath_linecount = 0
        current_subpath_linecount += 1
        total_linecount += 1

    if last_subpath is not None:
        mapping[last_subpath]["linecount"] = current_subpath_linecount
        mapping[last_subpath]["end"] = file.size

    file.length = total_linecount  # set the line count for further convenience
    return mapping



# pylint: disable=missing-docstring
import math

from resolwe.flow.utils import stats
from resolwe.test import TestCase


class TestStatUtils(TestCase):
    def test_shape_basic(self):
        series = stats.NumberSeriesShape()
        self.assertEqual(
            series.to_dict(),
            {
                "high": -math.inf,
                "low": math.inf,
                "mean": 0,
                "count": 0,
                "deviation": 0,
            },
        )

        series.update(1)
        self.assertEqual(series.count, 1)
        self.assertAlmostEqual(series.high, 1.0)
        self.assertAlmostEqual(series.low, 1.0)
        self.assertAlmostEqual(series.mean, 1.0)
        self.assertAlmostEqual(series.deviation, 0.0)

        for _ in range(5):
            series.update(1)
        self.assertEqual(series.count, 6)
        self.assertAlmostEqual(series.high, 1.0)
        self.assertAlmostEqual(series.low, 1.0)
        self.assertAlmostEqual(series.mean, 1.0)
        self.assertAlmostEqual(series.deviation, 0.0)

        large = 1000000.0
        series.update(large)
        series.update(-large)
        self.assertAlmostEqual(series.high, large)
        self.assertAlmostEqual(series.low, -large)
        self.assertAlmostEqual(series.mean, 0.75)
        self.assertAlmostEqual(series.deviation, 534522.483825049)

    def test_shape_tricky(self):
        series = stats.NumberSeriesShape()
        for _ in range(10):
            series.update(1)
        for _ in range(10):
            series.update(-1)
        self.assertAlmostEqual(series.mean, 0.0)
        self.assertAlmostEqual(series.deviation, 1.02597835208515)

    def test_load_avg_basic(self):
        avg = stats.SimpleLoadAvg([60])

        self.assertAlmostEqual(avg.intervals[60].value, 0.0)
        self.assertAlmostEqual(avg.intervals["1m"].value, 0.0)
        avg.add(1, timestamp=0.0)
        self.assertAlmostEqual(avg.intervals[60].value, 0.0)

        avg.add(1, timestamp=0.2)
        with self.assertRaises(ValueError):
            avg.add(1, timestamp=0.1)

        avg.add(1, timestamp=0.4)
        avg.add(1, timestamp=0.6)
        avg.add(1, timestamp=0.8)
        avg.add(1, timestamp=1.0)
        avg.add(1, timestamp=5)
        self.assertAlmostEqual(avg.intervals[60].value, 0.1, places=1)

    def test_load_avg_multi(self):
        avg = stats.SimpleLoadAvg([60, 5 * 60, 15 * 60])
        for i in range(1, 15 * 60):
            avg.add(1, timestamp=i)
        self.assertAlmostEqual(avg.intervals["1m"].value, 1, places=1)
        self.assertAlmostEqual(avg.intervals["5m"].value, 0.9, places=1)
        self.assertAlmostEqual(avg.intervals["15m"].value, 0.6, places=1)

    def test_load_avg_decay(self):
        avg = stats.SimpleLoadAvg([60])
        for i in range(0, 5 * 60):
            avg.add(2, timestamp=i)
        self.assertAlmostEqual(avg.intervals["1m"].value, 2, places=1)

        for i in range(5 * 60, 10 * 60):
            avg.add(5, timestamp=i)
        self.assertAlmostEqual(avg.intervals["1m"].value, 5, places=1)

        for i in range(10 * 60, 15 * 60):
            avg.add(0, timestamp=i)
        self.assertAlmostEqual(avg.intervals["1m"].value, 0, places=1)

    def test_load_avg_display(self):
        self.assertEqual(stats._display_interval(59), "59s")
        self.assertEqual(stats._display_interval(60), "1m")
        self.assertEqual(stats._display_interval(65), "1m5s")
        self.assertEqual(stats._display_interval(300), "5m")
        self.assertEqual(stats._display_interval(3666), "1h1m6s")
        self.assertEqual(stats._display_interval(260200), "3d16m40s")



# Generated by Django 3.1.7 on 2021-10-24 19:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    dependencies = [
        ("permissions", "0001_initial"),
        ("flow", "0002_squashed_0049_json_field"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="collection",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view collection"),
                    ("edit", "Can edit collection"),
                    ("share", "Can share collection"),
                    ("owner", "Is owner of the collection"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="data",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view data"),
                    ("edit", "Can edit data"),
                    ("share", "Can share data"),
                    ("owner", "Is owner of the data"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="descriptorschema",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view descriptor schema"),
                    ("edit", "Can edit descriptor schema"),
                    ("share", "Can share descriptor schema"),
                    ("owner", "Is owner of the description schema"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="entity",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view entity"),
                    ("edit", "Can edit entity"),
                    ("share", "Can share entity"),
                    ("owner", "Is owner of the entity"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="process",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view process"),
                    ("share", "Can share process"),
                    ("owner", "Is owner of the process"),
                ),
            },
        ),
        migrations.AlterModelOptions(
            name="relation",
            options={
                "default_permissions": (),
                "get_latest_by": "version",
                "permissions": (
                    ("view", "Can view relation"),
                    ("edit", "Can edit relation"),
                    ("share", "Can share relation"),
                    ("owner", "Is owner of the relation"),
                ),
            },
        ),
        migrations.AddField(
            model_name="collection",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="collection",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="data",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="data",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="descriptorschema",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="descriptorschema",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="entity",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="entity",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="process",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="process",
                to="permissions.permissiongroup",
            ),
        ),
        migrations.AddField(
            model_name="relation",
            name="permission_group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="relation",
                to="permissions.permissiongroup",
            ),
        ),
    ]



"""Resolwe models reference utils."""
from resolwe.flow.utils import iterate_fields


def referenced_schema_files(fields, schema):
    """Get the list of files and directories references by fields.

    :return: tuple of lists, first list containing  files and
        directories refereced in data.output.
    :rtype: Tuple[List[str], List[str]]
    """
    refs = []
    for field_schema, fields in iterate_fields(fields, schema):
        if "type" in field_schema:
            field_type = field_schema["type"]
            field_name = field_schema["name"]

            # Add basic:file: entries
            if field_type.startswith("basic:file:"):
                refs.append(fields[field_name]["file"])
                refs += fields[field_name].get("refs", [])

            # Add list:basic:file: entries
            elif field_type.startswith("list:basic:file:"):
                for field in fields[field_name]:
                    refs.append(field["file"])
                    refs += field.get("refs", [])

            # Add basic:dir: entries
            elif field_type.startswith("basic:dir:"):
                refs.append(fields[field_name]["dir"])
                refs += fields[field_name].get("refs", [])

            # Add list:basic:dir: entries
            elif field_type.startswith("list:basic:dir:"):
                for field in fields[field_name]:
                    refs.append(field["dir"])
                    refs += field.get("refs", [])
    return refs


def referenced_files(data, include_descriptor=True):
    """Get the list of files and directories referenced by the data object.

    :param data: given data object.
    :type data: resolwe.flow.models.Data

    :param include_descriptor: include files referenced in descriptor schema.
    :type include_descriptor: bool

    :return: referenced files and directories.
    :rtype: List[str]
    """
    special_files = ["jsonout.txt", "stdout.txt"]

    output = data.output
    output_schema = data.process.output_schema

    descriptor = data.descriptor
    descriptor_schema = getattr(data.descriptor_schema, "schema", [])

    output_files = referenced_schema_files(output, output_schema)

    descriptor_files = (
        referenced_schema_files(descriptor, descriptor_schema)
        if include_descriptor
        else []
    )

    return special_files + output_files + descriptor_files



""".. Ignore pydocstyle D400.

=======================
Interpret Test Profiles
=======================

"""
import json

from django.core.management.base import BaseCommand


class Command(BaseCommand):
    """Utility for interpreting test profiles."""

    help = "Utility for interpreting test profiles."

    def add_arguments(self, parser):
        """Add command arguments."""
        parser.add_argument("log", nargs="+", type=str, help="profile log filename")
        parser.add_argument(
            "--sort",
            type=str,
            choices=("name", "time"),
            default="name",
            help="sort order (default: name)",
        )

    def handle(self, log, **options):
        """Handle command."""
        results = {}
        for filename in log:
            with open(filename) as log_file:
                for line in log_file:
                    data = json.loads(line)
                    test = results.setdefault(data["test"], {})
                    del data["test"]
                    test.update(data)

        # Generate sort function.
        if options["sort"] == "name":

            def sort_key(item):
                """Sort by block name."""
                key, _ = item
                return key

            sort_reverse = False
        elif options["sort"] == "time":

            def sort_key(item):
                """Sort by block total time."""
                _, blocks = item
                return blocks["total"]

            sort_reverse = True

        totals = {}
        for test, blocks in sorted(results.items(), key=sort_key, reverse=sort_reverse):
            print("{}".format(test))

            total_time = blocks["total"]
            for block, time in blocks.items():
                totals[block] = totals.get(block, 0.0) + time
                print(
                    "  {}: {:.2f} ({:.2f}%)".format(
                        block, time, time * 100.0 / total_time
                    )
                )

            print("")

        print("Overall:")
        total_time = totals["total"]
        for block, time in totals.items():
            print(
                "  {}: {:.2f} ({:.2f}%)".format(block, time, time * 100.0 / total_time)
            )
